"use strict";

var toReference = require('./helpers/to-reference');
var flattenExpressions = require('./helpers/flatten-expressions');
var createTemplateKey = require('./createTemplateKey');
var addTemplatesToModule = require('./addTemplatesToModule');

function nullObject() {
  return Object.create(null);
}

function setupInjector(program, parent, scope, file) {
	file.setDynamic(namespace, nullObject);
}

function isComponent(name) {
	return name.charAt(0).toUpperCase() === name.charAt(0);
}

//skip passing them as values
var literalTypes = {
	Literal: true,
	StringLiteral: true,
	NumericLiteral: true,
	BooleanLiteral: true
};

function handleWhiteSpace(str) {
	str = str.replace(/\t/g, '');
	if(str.indexOf("\n") !== -1) {
		str = str.substring(0, str.indexOf("\n")).trim();
	}
	return str;
}

function processAttributes(t, attributes, root, templateElem, isComponent) {
	var props = [];

	for(var i = 0; i < attributes.length; i++) {
		var attribute = attributes[i];

		if(attribute.type === 'JSXAttribute') {
			var attrName = attribute.name.name;
			var expression = attribute.value.expression;

			if(expression !== undefined) {
				if (literalTypes[expression.type]) {
					root.templateString += expression.value + "|-|";
					if(!isComponent) {
						templateElem.attrs[attrName] = expression.value;
					} else {
						props.push(
							t.ObjectProperty(t.identifier(attrName), expression)
						);
					}
				} else if (expression.type === "NullLiteral") {
					root.templateString += "null|-|";
					if(!isComponent) {
						templateElem.attrs[attrName] = null;
					} else {
						props.push(
							t.ObjectProperty(t.identifier(attrName), expression)
						);
					}
				} else if (expression.type === "BinaryExpression"
					&& literalTypes[expression.left.type] && literalTypes[expression.right.type]) {

					root.templateString += expression.left.value + " " + expression.operator + " " + expression.right.value +  "|-|";
					if(!isComponent) {
						templateElem.attrs[attrName] = expression;
					} else {
						props.push(
							t.ObjectProperty(t.identifier(attrName), expression)
						);
					}
				} else {
					var index = root.templateValues.length;

					root.templateString += "$$|";
					root.expressionMap[expression] = index;
					if(!isComponent) {
						root.templateValues.push(expression);
						templateElem.attrs[attrName] = {
							index: index
						};
					} else {
						props.push(
							t.ObjectProperty(t.identifier(attrName), expression)
						);
					}
				}
			} else if (literalTypes[attribute.value.type]) {
				root.templateString += attribute.value.value + "|-|";
				if(!isComponent) {
					templateElem.attrs[attrName] = attribute.value.value;
				} else {
					props.push(
						t.ObjectProperty(t.identifier(attrName), t.StringLiteral(attribute.value.value))
					);
				}
			} else if (attribute.value.type === "NullLiteral") {
				root.templateString += "null|-|";
				if(!isComponent) {
					templateElem.attrs[attrName] = null;
				} else {
					props.push(
						t.ObjectProperty(t.identifier(attrName), t.identifier('null'))
					);
				}
			}
		}
	}

	if(isComponent) {
		return t.ObjectExpression(props);
	}
}

function processElement(t, element, root, parentTemplateElem) {
	if (element.type === "JSXElement") {
		if (element.openingElement) {
			var tagName = element.openingElement.name.name;
			var templateElem;

			if (!isComponent(tagName)) {
				templateElem = {
					tag: tagName,
					children: null,
					attrs: {}
				};

				if(element.openingElement.attributes && element.openingElement.attributes.length > 0) {
					processAttributes(t, element.openingElement.attributes, root, templateElem, false);
				}

			} else {
				var index = root.templateValues.length;
				root.expressionMap[tagName] = index;

				templateElem = {
					component: {
						index: index
					},
					children: null
				};

				var props = null;

				if(element.openingElement.attributes && element.openingElement.attributes.length > 0) {
					props = processAttributes(t, element.openingElement.attributes, root, templateElem, true);
				}

				root.templateValues.push(
					t.ObjectExpression([
						t.ObjectProperty(t.identifier('component'), t.identifier(tagName)),
						t.ObjectProperty(t.identifier('props'), props !== null ? props : t.identifier('null'))
					])
				);
			}

			root.templateString += tagName + "|";
			if (!root.templateElem) {
				root.templateElem = templateElem;
			}
			if (!element.selfClosing) {
				templateElem.children = [];
				processChildren(t, element.children, root, templateElem);
			}
			if (parentTemplateElem) {
				parentTemplateElem.children.push(templateElem);
			}
		}
	} else if (element.type === "JSXExpressionContainer") {
		var index = root.templateValues.length;

		root.templateString += "$$|";
		root.expressionMap[element.expression] = index;
		root.templateValues.push(element.expression);
		parentTemplateElem.children.push({
			index: index
		});
	} else if (element.type === "Literal") {
		parentTemplateElem.children.push(element.value);
		root.templateValues.push(element);
		root.templateString += element.value + "|";
	} else if (element.type === "JSXText") {
		var text = handleWhiteSpace(element.value);
		if(text != '') {
			parentTemplateElem.children.push(text);
			root.templateString += text + "|";
		}
	}
}

function processChildren(t, children, root, parentTemplateElem) {
	if(children) {
		for(var i = 0; i < children.length; i++) {
			var child = children[i];
			processElement(t, child, root, parentTemplateElem);
		}
	}
}

module.exports = function(options) {
	var t = options.types;

	return {
		visitor : {
			Program: {
				exit(path, scope) {
					var opts = scope.opts;
					var node = path.node;

					for(var templateKey in opts.roots) {
						var root = opts.roots[templateKey];
						addTemplatesToModule(t, node, templateKey, root);
					}
				}
			},
			JSXElement: {
				enter(path, scope) {
					var opts = scope.opts;
					var node = path.node;

					if (node.root !== undefined) {
						return;
					}

					var root = {
						templateValues: [],
						templateElem: null,
						templateString: "",
						templateKey: null,
						expressionMap: {}
					};

					processElement(t, node, root, null);
					//create the templateKey
					root.templateKey =  "tpl" + createTemplateKey(root.templateString);

					var values = t.identifier('null');
					var expressions = flattenExpressions(t, root.templateValues);

					if (root.templateValues.length === 1) {
						values = t.toExpression(expressions[0]);
					} else if(root.templateValues.length > 1) {
						values = t.arrayExpression(expressions);
					}

					if (!opts.roots) {
						opts.roots = {};
					}
					opts.roots[root.templateKey] = root;

					path.replaceWith(
						t.ExpressionStatement(t.callExpression(
							t.memberExpression(t.identifier("Inferno"), t.identifier("createFragment")),
							[values, t.identifier(root.templateKey)]
						))
					);
				}
			}
		},
		inherits: require("babel-plugin-syntax-jsx")
	}
};

